'use strict';
const co = require('co');
const fs = require('fs');
const path = require('path');

const FLAG_INJECT = '$inject';
const FLAG_BEFORE = '$before';
const FLAG_AFTER = '$after';

function isGenerator(obj) {
    return 'function' == typeof obj.next && 'function' == typeof obj.throw;
}

function isGeneratorFunction(obj) {
    var constructor = obj.constructor;
    if (!constructor) return false;
    if ('GeneratorFunction' === constructor.name || 'GeneratorFunction' === constructor.displayName) return true;
    return isGenerator(constructor.prototype);
}

class Manager {
    constructor(options) {
        this.options = options || {};
        this.actions = {};
    }

    /**
     * 指空目录下的所有Action,加载到管理器内.
     *
     * @param {String} [actionDir]             - 文件目录
     * @returns {Manager}
     * */
    load(actionsDir) {
        let dir = actionsDir || this.options.dir || path.join(process.cwd(), '/actions');
        let files = [];
        if (fs.existsSync(dir)) {
            files = fs.readdirSync(dir);
        }
        files.forEach((file)=> {
            let action = require(path.join(dir, file));
            this.add(path.basename(file, '.js'), action);
        });

        return this;
    }

    /**
     * 向管理器添加Action
     *
     * @param {String} name                        - Action名称
     * @param {Function|GeneratorFunction} action  - Action
     * @return {Manager}
     * */
    add(name, action) {
        if (typeof action !== 'function') {
            return;
        }

        let funs = [];
        let beforeFuns = action[FLAG_BEFORE];
        let afterFuns = action[FLAG_AFTER];

        //action before
        Array.isArray(beforeFuns) && beforeFuns.forEach(fun => funs.push(fun));
        //action 是个GeneratorFunction使用co包成普通的函数(返回promise)
        isGeneratorFunction(action) ? funs.push(co.wrap(action)) : funs.push(action);
        //action after
        Array.isArray(afterFuns) && afterFuns.forEach(fun => funs.push(fun));

        this.actions[name] = {inject: action[FLAG_INJECT] || [], funs: funs};

        return this;
    }

    /**
     * 移除管理器内,指定Action
     *
     * @param {String} name             - Action名称
     * @return {Manager}
     * */
    remove(name) {
        delete this.actions[name];
        return this;
    }

    /**
     * 运行管理器内,指定的Action
     *
     * @param {String} name             - Action名称
     * @param {Array} [argv]            - Action执行时参数
     * @param {Object} [container]      - Action执行时的容器. 注入的对象需要从此容器内获取
     *
     * @returns {Promise}
     * */
    run(name, argv, container) {
        let action = this.actions[name];
        let context = {};

        try {
            if (!action) {
                throw new Error(`[action:${name}] 不存在`)
            }
            action.inject.forEach(injectName => {
                let injectObj = (container || {})[injectName];
                if (!injectObj) {
                    throw new Error(`[action:${name}] 需要注入的对象[${injectName}]在容器内不存在`);
                }
                Object.defineProperty(context, injectName, {
                    configurable: false,
                    enumerable: false,
                    writable: false,
                    value: Object.seal(injectObj)       //注入的对象不能被添加&删除任何属性
                });
            });
        } catch (error) {
            return Promise.reject(error);
        }

        return new Promise((resolve, reject)=> {
            let funs = action.funs;
            let i = -1;
            let fun;
            let promise;
            next();
            function next() {
                fun = funs[++i];
                if (!fun) {         //completed
                    resolve(context);
                    return;
                }
                try {
                    promise = fun.apply(context, argv);
                    if (!(promise instanceof Promise)) {
                        promise = Promise.resolve(promise);
                    }
                    promise.then((data) => next(data)).catch(reject);
                } catch (error) {
                    reject(error);
                }
            }
        });
    }
}

module.exports.Manager = Manager;
module.exports.create = function (options) {
    return new Manager(options);
};

